"""wsdl4py

This library can be used to read WSDL messages into a python class structure.  You can also create
new instances of the WSDL class to define new messages.

Example:  Create form a URI

wsdl = ReadFromUri('http://fourthought.com/defs.wsdl')

Example:  Create a new WSDL definition

from xml.dom.ext import PrettyPrint
wsdl = Wsdl()
m = wsdl.addMessage('messageName')
m.documentation = 'Documentation about the message'
d = wsdl.toDom()
PrettyPrint(d)

For more information see the IBM developerWorks tutorial on wsdl4py at

http://www-105.ibm.com/developerworks/education.nsf/webservices-onlinecourse-bytitle/D092605075D1D67D86256A7E0042432D?OpenDocument

(free registration required)
"""

import urllib, UserDict, cStringIO
import os

# Release 0.12 of 4Suite unified the domelettes and removed pDomlette
DOMLETTE_UNIFICATION = False
try:
    import Ft.Xml.Domlette
    DOMLETTE_UNIFICATION = True
except ImportError:
   from Ft.Lib import pDomlette

if DOMLETTE_UNIFICATION:
    from Ft.Xml import XPath
    xpath = XPath
    from Ft.Xml.XPath import Conversions
    PrettyPrint = Ft.Xml.Domlette.PrettyPrint
    from Ft.Rdf.Inference.Common import SplitQName
    import xml.dom
    EMPTY_NAMESPACE = xml.dom.EMPTY_NAMESPACE
else:
    from xml import xpath
    from xml.xpath import Conversions
    from xml.dom.ext import PrettyPrint, SplitQName
    EMPTY_NAMESPACE = ""

from xml.dom import Node

WSDL_NAMESPACE='http://schemas.xmlsoap.org/wsdl/'


#These expressions are used to pull information from a XML document
g_nameExpression = xpath.Compile('string(@name)')
g_targetNamespaceExpression = xpath.Compile('string(@targetNamespace)')
g_importExpression = xpath.Compile('w:import')
g_documentationExpression = xpath.Compile('w:documentation')
g_typesExpression = xpath.Compile('w:types')
g_messagesExpression = xpath.Compile('w:message')
g_portTypesExpression = xpath.Compile('w:portType')
g_bindingsExpression = xpath.Compile('w:binding')
g_servicesExpression = xpath.Compile('w:service')
g_partsExpression = xpath.Compile('w:part')
g_operationsExpression = xpath.Compile('w:operation')
g_inputExpression = xpath.Compile('w:input')
g_outputExpression = xpath.Compile('w:output')
g_faultExpression = xpath.Compile('w:fault')
g_portsExpression = xpath.Compile('w:port')

class WsdlException(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg


def ReadFromStream(stream, uri='file://' + os.getcwd() + os.sep + 'foo'):
    """Read in a WSDL message from a steam object
    
    4Suite 0.12 requires a URI when parsing xml data, as it is required to
    resolve XIncludes and entities.   It is also required for some XSLT (for
    example xsl:include and xsl:import) and RDF."""

    ws = Wsdl()
    if DOMLETTE_UNIFICATION:
        dom = Ft.Xml.Domlette.NonvalidatingReader.parseStream(stream, uri)
        ws.fromDom(dom)
    else:
        reader = pDomlette.PyExpatReader()
        dom = reader.fromStream(stream)
        ws.fromDom(dom)
        reader.releaseNode(dom)
    return ws


def ReadFromString(st):
    """Read in a WSDL message from a string"""
    st = cStringIO.StringIO(st)
    return ReadFromStream(st)

def ReadFromUri(uri):
    """Read in a WSDL message from a URI"""
    st = urllib.urlopen(uri)
    ws = ReadFromStream(st, uri)
    st.close()
    return ws

def getAttributeNSWrapper(node, namespaceprefix, attribute):
    """Newer stuff uses xml.dom.EMPTY_NAMESPACE (None) instead of ''."""

    if not namespaceprefix and DOMLETTE_UNIFICATION:
        return node.getAttributeNS(xml.dom.EMPTY_NAMESPACE, attribute)
    else:
        return node.getAttributeNS(namespaceprefix, attribute)

class Wsdl:
    """This class represents the entire WSDL definition.  It can be created from one of the above
    constructor functions, or a new instance can be created and the attributes of the class set
    manually."""

    def __init__(self):
        self.types = Types()
        self.messages = Messages()
        self.portTypes = PortTypes()
        self.bindings = Bindings()
        self.services = Services()

        self.name = ''
        self.targetNamespace = ''
        self.documentation = ''


    def fromDom(self,dom):
        """Read in a WSDL definition from DOM.  This is called by the Read* methods at the top of the module"""

        con = xpath.Context.Context(dom.documentElement, processorNss={'w': WSDL_NAMESPACE})

        self.setName(g_nameExpression.evaluate(con))
        self.setTargetNamespace(g_targetNamespaceExpression.evaluate(con))

        imports = g_importExpression.evaluate(con)
        for i in imports:
            raise Exception("Imports not supported")

        docs = g_documentationExpression.evaluate(con)
        if docs:
            self.setDocumentation(Conversions.StringValue(docs[0]))

        types = g_typesExpression.evaluate(con)
        self.types.fromDom(types)

        messages = g_messagesExpression.evaluate(con)
        self.messages.fromDom(messages)

        portTypes = g_portTypesExpression.evaluate(con)
        self.portTypes.fromDom(portTypes)

        bindings = g_bindingsExpression.evaluate(con)
        self.bindings.fromDom(bindings)

        services = g_servicesExpression.evaluate(con)
        self.services.fromDom(services)
    

    def setName(self,newName):
        """Change the name of the definition"""
        self.name = newName

    def setTargetNamespace(self,newTns):
        """Change the target namespace of the definition"""
        self.targetNamespace = newTns

    def setDocumentation(self,documentation):
        self.documentation = documentation

    def toDom(self):
        """Generate a DOM document for the current WSDL definition structure"""
        if DOMLETTE_UNIFICATION:
            doc = Ft.Xml.Domlette.implementation.createDocument(
                WSDL_NAMESPACE, 'wsdl:definitions', None
                )
            defs = doc.documentElement
        else:
            doc = pDomlette.Document()
            defs = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:definitions')

        if self.name:
            defs.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)
        if self.targetNamespace:
            defs.setAttributeNS(EMPTY_NAMESPACE, 'targetNamespace', self.targetNamespace)

        DocumentationToDom(defs,self.documentation)

        doc.appendChild(defs)

        self.types.toDom(doc)
        self.messages.toDom(doc)
        self.portTypes.toDom(doc)
        self.bindings.toDom(doc)
        self.services.toDom(doc)

        return doc
        

    def validate(self):
        """Validate the current WSDL structure.  Perform checks that all of the interconnected parts (messages,
        portTypes, services, etc) reference valid definitions"""


        #FIXME We cannot validate messages to definitions because we are unable to parse the definition

        self.portTypes.validate(self.messages)
        self.bindings.validate(self.portTypes)
        #self.services.validate(self.bindings)


class Types:
    """A representation of the type definitions in a WSDL definition.  This will some day be
    incorporated into a schema parser that can validate that the type definitions are valid.  Currently
    it stores the strings representation of the type definition."""

    def __init__(self):
        self.documentation = ""
        self.definitions = []

    def fromDom(self,typeNodes):
        """Create a new type definition from a DOM.  Generally called by Wsdl.fromDom"""

        if not len(typeNodes):
            return
        types = typeNodes[0]
        con = xpath.Context.Context(types, processorNss={'w': WSDL_NAMESPACE})
        docs = g_documentationExpression.evaluate(con)
        if docs:
            self.setDocumentation(Conversions.StringValue(docs[0]))

        for child in types.childNodes:
            if child.nodeType != Node.ELEMENT_NODE:
                continue
            self.definitions.append(ExtensionFromDom(child))

    def toDom(self,doc):
        """Generate the DOM equivalent for this type definition.  Generally called by Wsdl.toDom"""

        de = doc.documentElement
        types = doc.createElementNS(WSDL_NAMESPACE,'wsdl:types')
        DocumentationToDom(types,self.documentation)
        ExtensionsToDom(types,self.definitions)
        de.appendChild(types)


class Messages(UserDict.UserDict):

    def addMessage(self,name,documentation = ""):
        """Add a new message to the WSDL definition"""
        if self.has_key(name):
            raise WsdlException("Message %s already exists" % name)
        m = Message(name)
        m.documentation = documentation
        self[name] = m
        return m

    def fromDom(self,messages):
        for m in messages:
            con = xpath.Context.Context(m, processorNss={'w': WSDL_NAMESPACE})
            documentation = DocumentationFromDom(con)
            msg = self.addMessage(getAttributeNSWrapper(m, '','name'),
                    documentation)

            parts = g_partsExpression.evaluate(con)
            for p in parts:
                n = getAttributeNSWrapper(p,'','name')
                element = getAttributeNSWrapper(p,'','element')
                _type = getAttributeNSWrapper(p,'','type')
                msg.addPart(n,element,_type)            

    def toDom(self,doc):
        for message in self.values():
            message.toDom(doc)

class Message:
    def __init__(self,name):
        self.name = name
        self.documentation = ""
        self.parts = {}


    def addPart(self,name,element='',_type=''):
        """Add a new part to a message"""
        if self.parts.has_key(name):
            raise WsdlException("Message '%s' already has part '%s' exists" \
                    % (self.name,name))
        p = Part(name)
        p.element = element
        p.type = _type
        self.parts[name] = p
        return p
        

    def toDom(self, doc):
        de = doc.documentElement
        m = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:message')
        m.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)
        DocumentationToDom(m,self.documentation)
        for p in self.parts.values():
            p.toDom(m)
        de.appendChild(m)


class Part:
    def __init__(self,name):
        self.name = name
        self.element = ""
        self.type = ""

    def toDom(self,message):
        doc = message.ownerDocument
        p = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:part')
        p.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)
        if self.element:
            p.setAttributeNS(EMPTY_NAMESPACE, 'element', self.element)
        if self.type:
            p.setAttributeNS(EMPTY_NAMESPACE, 'type', self.type)
        message.appendChild(p)
        

class PortTypes(UserDict.UserDict):

    def addPortType(self,name,documentation=''):
        """Add a new port type to the definition"""
        if self.has_key(name):
            raise WsdlException("Port Type %s already exists" % (name))
        pt = PortType(name)
        pt.documentation = documentation
        self[name] = pt
        return pt

    def fromDom(self,portTypes):

        for pt in portTypes:
            if pt.nodeType != Node.ELEMENT_NODE:
                continue
            con = xpath.Context.Context(pt, processorNss={'w': WSDL_NAMESPACE})

            documentation = DocumentationFromDom(con)
            p = self.addPortType(getAttributeNSWrapper(pt,'','name'),documentation)

            operations = g_operationsExpression.evaluate(con)
            for op in operations:
                con = xpath.Context.Context(op, processorNss={'w': WSDL_NAMESPACE})
                documentation = DocumentationFromDom(con)
                o = p.addOperation(getAttributeNSWrapper(op,'','name'),documentation)
                self.__inputFromDom(con,o)
                self.__outputFromDom(con,o)
                self.__faultsFromDom(con,o)

    def toDom(self,doc):
        for pt in self.values():
            pt.toDom(doc)


    def validate(self,messages):
        """Make sure all of our operations have valide inputs and outputs"""
        for pt in self.values():
            pt.validate(messages)
        

    def __inputFromDom(self,con,o):
        input = g_inputExpression.evaluate(con)
        if len(input):
            input = input[0]
            msg = getAttributeNSWrapper(input,'','message')
            name = getAttributeNSWrapper(input,'','name')
            icon = xpath.Context.Context(input, processorNss={'w': WSDL_NAMESPACE})
            documentation = DocumentationFromDom(icon)
            o.setInput(msg,name,documentation)

    def __outputFromDom(self,con,o):
        output = g_outputExpression.evaluate(con)
        if len(output):
            output = output[0]
            msg = getAttributeNSWrapper(output,'','message')
            name = getAttributeNSWrapper(output,'','name')
            ocon = xpath.Context.Context(output, processorNss={'w': WSDL_NAMESPACE})
            documentation = DocumentationFromDom(ocon)
            o.setOutput(msg,name,documentation)

    def __faultsFromDom(self,con,o):
        faults = g_faultExpression.evaluate(con)
        for fault in faults:
            msg = getAttributeNSWrapper(fault,'','message')
            name = getAttributeNSWrapper(fault,'','name')
            fcon = xpath.Context.Context(fault, processorNss={'w': WSDL_NAMESPACE})
            documentation = DocumentationFromDom(fcon)
            o.addFault(name,msg,documentation)


        

class PortType:

    def __init__(self,name):
        self.name = name
        self.documentation = ""
        self.operations = {}


    def addOperation(self,name,documentation=''):
        """Add a new operation to a port type"""
        if self.operations.has_key(name):
            raise WsdlException("Port Type %s already has operation %s" % (self.name,name))
        o = Operation(name)
        o.documentation = documentation
        self.operations[name] = o
        return o
        

    def toDom(self, doc):
        de = doc.documentElement

        pt = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:portType')
        pt.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)
        DocumentationToDom(pt,self.documentation)
        for op in self.operations.values():
            op.toDom(pt)
        
        de.appendChild(pt)

    def validate(self,messages):
        """Make sure all of our operations have valid inputs, outputs and faults"""
        for op in self.operations.values():
            op.validate(messages)
        


class Operation:

    def __init__(self,name):
        self.name = name
        self.documentation = ""
        self.input = None
        self.output = None
        self.faults = {}

    def addFault(self,name,message,documentation=''):
        """Add a new fault to the operation"""
        if self.faults.has_key(name):
            raise WsdlException("Operation %s already has fault %s" % (self.name,name))

        f = Fault(name,message)
        f.documentation = documentation
        self.faults[name] = f
        return f


    def setInput(self,message,name='',documentation=''):
        i = Input(message)
        i.name = name
        i.documentation = documentation
        self.input = i
        return i

    def setOutput(self,message,name='',documentation=''):
        o = Output(message)
        o.name = name
        o.documentation = documentation
        self.output = o
        return o


    def toDom(self,portType):
        doc = portType.ownerDocument

        op = doc.createElementNS(WSDL_NAMESPACE,'wsdl:operation')
        op.setAttributeNS(EMPTY_NAMESPACE,'name',self.name)

        DocumentationToDom(op,self.documentation)

        if self.input:
            self.input.toDom(op)
        if self.output:
            self.output.toDom(op)
        for f in self.faults.values():
            f.toDom(op)

        portType.appendChild(op)

    def validate(self,messages):
        """Make sure our input, output and faults refrence valid messages"""
        if self.input and not ValidateTns(self.input.message,messages):
            raise WsdlValidationException("Operation %s, Output references invalid message %s" % (self.name,
                                                                                                  self.input.message))

        if self.output and not ValidateTns(self.output.message,messages):
            raise WsdlValidationException("Operation %s, Output references invalid message %s" % (self.name,
                                                                                                  self.output.message))

        for fault in self.faults.values():
            if not ValidateTns(fault.message,messages):
                raise WsdlValidationException("Operation %s, Fault %s references invalid message %s" % (self.name,
                                                                                                        fault.name,
                                                                                                        fault.message))


class Input:

    def __init__(self,message):
        self.name = ''
        self.message = message
        self.documentation = ''

    def toDom(self,operation):
        doc = operation.ownerDocument
        i = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:input')
        i.setAttributeNS(EMPTY_NAMESPACE, 'message', self.message)
        if self.name:
            i.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)
            
        DocumentationToDom(i,self.documentation)
        operation.appendChild(i)

class Output:

    def __init__(self,message):
        self.name = ''
        self.message = message
        self.documentation = ''

    def toDom(self,operation):
        doc = operation.ownerDocument
        o = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:output')
        o.setAttributeNS(EMPTY_NAMESPACE, 'message', self.message)
        if self.name:
            o.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)
            
        DocumentationToDom(o,self.documentation)
        operation.appendChild(o)


class Fault:

    def __init__(self,name,message):
        self.name = name
        self.message = message
        self.documentation = ''

    def toDom(self,operation):
        doc = operation.ownerDocument
        f = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:fault')
        f.setAttributeNS(EMPTY_NAMESPACE, 'message', self.message)
        if self.name:
            f.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)
            
        DocumentationToDom(f,self.documentation)
        operation.appendChild(f)

class Bindings(UserDict.UserDict):


    def addBinding(self,name,_type,documentation=''):
        """Add a new binding to the definition"""
        if self.has_key(name):
            raise WsdlException("fault %s alread exists" % (name))

        b = Binding(name,_type)
        b.documentation = documentation
        self[name] = b
        return b 

    def fromDom(self,bindings):
        
        for b in bindings:
            if b.nodeType != Node.ELEMENT_NODE:
                continue


            con = xpath.Context.Context(b, processorNss={'w': WSDL_NAMESPACE})

            documentation = DocumentationFromDom(con)
            bi = self.addBinding(getAttributeNSWrapper(b,'','name'),
                    getAttributeNSWrapper(b,'','type'),documentation)

            for child in b.childNodes:
                if child.nodeType != Node.ELEMENT_NODE:
                    continue
                if child.namespaceURI != WSDL_NAMESPACE or child.localName != 'operation':
                    #An extension Element
                    bi.addExtension(ExtensionFromDom(child))
                else:
                    con = xpath.Context.Context(child, processorNss={'w': WSDL_NAMESPACE})

                    documentation = DocumentationFromDom(con)
                    o = bi.addOperation(getAttributeNSWrapper(child,'','name'),
                            documentation)

                    self.__readInputFromDom(con,o)
                    self.__readOutputFromDom(con,o)
                    self.__readFaultsFromDom(con,o)            

    def toDom(self,doc):
        for binding in self.values():
            binding.toDom(doc)

    def validate(self,portTypes):
        for binding in self.values():
            binding.validate(portTypes)

    def __readInputFromDom(self,con,op):
        input = g_inputExpression.evaluate(con)
        if len(input):
            input = input[0]
            documentation = ''
            extensions = []
            for iChild in input.childNodes:
                if iChild.nodeType != Node.ELEMENT_NODE:
                    continue
                if (iChild.namespaceURI,iChild.localName) == (WSDL_NAMESPACE,'documentation'):
                    documentation = iChild.firstChild.data
                else:
                    #An extension
                    extensions.append(ExtensionFromDom(iChild))
            i = op.setInput(documentation)
            for e in extensions:
                i.addExtension(e)


    def __readOutputFromDom(self,con,op):
        output = g_outputExpression.evaluate(con)
        if len(output):
            output = output[0]
            documentation = ''
            extensions = []
            for iChild in output.childNodes:
                if iChild.nodeType != Node.ELEMENT_NODE:
                    continue
                if (iChild.namespaceURI,iChild.localName) == (WSDL_NAMESPACE,'documentation'):
                    documentation = iChild.firstChild.data
                else:
                    #An extension
                    extensions.append(ExtensionFromDom(iChild))

            bo = op.setOutput(documentation)
            for e in extensions:
                bo.addExtension(e)



    def __readFaultsFromDom(self,con,op):
        faults = g_faultExpression.evaluate(con)
        for fault in faults:
            documentation = ''
            extensions = []
            for iChild in fault.childNodes:
                if iChild.nodeType != Node.ELEMENT_NODE:
                    continue
                if (iChild.namespaceURI,iChild.localName) == (WSDL_NAMESPACE,'documentation'):
                    documentation = iChild.firstChild.data
                else:
                    #An extension
                    extensions.append(ExtensionFromDom(iChild))
            bf = op.addFault(getAttributeNSWrapper(fault,'','name'),documentation)
            for e in extensions:
                bf.addExtension(e)


class Binding:

    def __init__(self,name,_type):
        self.name = name
        self.type = _type
        self.extensions = []
        self.operations = {}

    def addOperation(self,name,documentation=''):
        if self.operations.has_key(name):
            raise WsdlException("operation %s alread exists on binding %s" % (name,self.name))

        o = BOperation(name)
        o.documentation = documentation
        self.operations[name] = o
        return o
        
    def addExtension(self,st):
        self.extensions.append(st)

    def toDom(self,doc):
        de = doc.documentElement
        b = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:binding')

        b.setAttributeNS(EMPTY_NAMESPACE,'name', self.name)
        b.setAttributeNS(EMPTY_NAMESPACE,'type', self.type)

        ExtensionsToDom(b,self.extensions)
        for op in self.operations.values():
            op.toDom(b)

        de.appendChild(b)

    def validate(self,portTypes):
        """Validate that this binding references a valid port type"""
        if not ValidateTns(self.type,portTypes):
            raise WsdlValidationException("Binding %s, references invalid portType %s" % (self.name,
                                                                                          self.type))

class BOperation:

    def __init__(self,name):
        self.name = name
        self.documentation = ""

        self.input = None
        self.output = None
        self.faults = {}

    def setInput(self,documentation=''):
        i = BInput()
        i.documentation = documentation
        self.input = i
        return i

    def setOutput(self,documentation=''):
        o = BOutput()
        o.documentation = documentation
        self.output = o
        return o

    def addFault(self,name,documentation=''):
        if self.faults.has_key(name):
            raise WsdlException("fault %s already exists on operation %s" % (name,self.name))
        f = BFault(name)
        f.documentation = documentation
        self.faults[name] = f
        return f

    def toDom(self,binding):
        doc = binding.ownerDocument

        o = doc.createElementNS(WSDL_NAMESPACE,'wsdl:operation')
        o.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)

        DocumentationToDom(o,self.documentation)

        if self.input:
            self.input.toDom(o)

        if self.output:
            self.output.toDom(o)

        for f in self.faults.values():
            f.toDom(o)


        binding.appendChild(o)


class BInput:

    def __init__(self):
        self.documentation = ''
        self.extensions = []
    
    def addExtension(self,ext):
        self.extensions.append(ext)
        
    def toDom(self,operation):
        doc = operation.ownerDocument

        i = doc.createElementNS(WSDL_NAMESPACE,'wsdl:input')

        DocumentationToDom(i,self.documentation)
        ExtensionsToDom(i,self.extensions)

        operation.appendChild(i)


class BOutput:

    def __init__(self):
        self.documentation = ''
        self.extensions = []
    
    def addExtension(self,ext):
        self.extensions.append(ext)
        
    def toDom(self,operation):
        doc = operation.ownerDocument

        o = doc.createElementNS(WSDL_NAMESPACE,'wsdl:output')

        DocumentationToDom(o,self.documentation)
        ExtensionsToDom(o,self.extensions)

        operation.appendChild(o)


class BFault:

    def __init__(self,name):
        self.name = name
        self.documentation = ''
        self.extensions = []
    
    def addExtension(self,ext):
        self.extensions.append(ext)
        
    def toDom(self,operation):
        doc = operation.ownerDocument

        f = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:fault')
        f.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)

        DocumentationToDom(f,self.documentation)
        ExtensionsToDom(f,self.extensions)

        operation.appendChild(f)


class Services(UserDict.UserDict):

    def addService(self,name,documentation=''):
        if self.has_key(name):
            raise WsdlException("service %s already exists" % (name))
        s = Service(name)
        s.documentation = documentation
        self[name] = s
        return s


    def fromDom(self,services):

        for serv in services:
            if serv.nodeType != Node.ELEMENT_NODE:
                continue

            con = xpath.Context.Context(serv, processorNss={'w': WSDL_NAMESPACE})
            documentation = DocumentationFromDom(con)
            s = self.addService(getAttributeNSWrapper(serv,'','name'),documentation)

            for pd in g_portsExpression.evaluate(con):
                pcon = xpath.Context.Context(pd, processorNss={'w': WSDL_NAMESPACE})
                documentation = DocumentationFromDom(pcon)
                p = s.addPort(getAttributeNSWrapper(pd,'','name'),
                        getAttributeNSWrapper(pd,'','binding'),documentation)

                for child in pd.childNodes:
                    if child.nodeType != Node.ELEMENT_NODE:
                        continue
                    if (child.namespaceURI,child.localName)  == (WSDL_NAMESPACE,'documentation'):
                        continue
                    else:
                        #An extension
                        p.addExtension(ExtensionFromDom(child))


            for child in serv.childNodes:
                if child.nodeType != Node.ELEMENT_NODE:
                    continue
                if (child.namespaceURI,child.localName) in [(WSDL_NAMESPACE,'documentation'),
                                                              (WSDL_NAMESPACE,'port')]:
                    continue
                else:
                    #An extension
                    s.addExtension(ExtensionFromDom(child))


    def toDom(self,doc):
        for s in self.values():
            s.toDom(doc)


class Service:
    def __init__(self,name):
        self.name = name
        self.documentation = ""
        self.ports = {}
        self.extensions = []

    def addPort(self, name, binding, documentation=''):
        if self.ports.has_key(name):
            raise WsdlException("port %s already exists on service %s" % (name,self.name))
        p = Port(name,binding)
        p.documentation = documentation
        self.ports[name] = p
        return p

    def addExtension(self,ext):
        self.extensions.append(ext)

    def toDom(self,doc):
        de = doc.documentElement
        s = doc.createElementNS(WSDL_NAMESPACE, 'wsdl:service')
        s.setAttributeNS(EMPTY_NAMESPACE, 'name', self.name)
        DocumentationToDom(s,self.documentation)

        for p in self.ports.values():
            p.toDom(s)

        ExtensionsToDom(s,self.extensions)

        de.appendChild(s)


class Port:

    def __init__(self,name,binding):
        self.name = name
        self.binding = binding
        self.documentation = ""
        self.extensions = []

    def addExtension(self,ext):
        self.extensions.append(ext)

    def toDom(self,port):
        doc = port.ownerDocument
        p = doc.createElementNS(WSDL_NAMESPACE,'wsdl:port')
        p.setAttributeNS(EMPTY_NAMESPACE,'name',self.name)
        p.setAttributeNS(EMPTY_NAMESPACE,'binding',self.binding)
        ExtensionsToDom(p,self.extensions)

        port.appendChild(p)

########
#Utility functions
########
def DocumentationToDom(parent,doc):
    if doc:
        dc = parent.ownerDocument.createElementNS(WSDL_NAMESPACE,'wsdl:documentation')
        dc.appendChild(parent.ownerDocument.createTextNode(doc))
        parent.appendChild(dc)
    

def DocumentationFromDom(con):
    docs = g_documentationExpression.evaluate(con)
    if docs:
        return Conversions.StringValue(docs[0])
    return ''

def ExtensionFromDom(node):
    st = cStringIO.StringIO()
    PrettyPrint(node,stream=st)
    return st.getvalue()
    
def ExtensionsToDom(parent, exts):
    reader = None
    if not DOMLETTE_UNIFICATION:
        reader = pDomlette.PyExpatReader()
    for e in exts:
        if DOMLETTE_UNIFICATION:
            dom = Ft.Xml.Domlette.NonvalidatingReader.parseString(e)
        else:
            dom = reader.fromString(e)
        parent.appendChild(dom.documentElement)
        if not DOMLETTE_UNIFICATION:
            reader.releaseNode(dom)
            
def ValidateTns(qname,sources):
    prefix,localName = SplitQName(qname)
    if prefix == 'tns':
        #Local to this document
        if not sources.has_key(localName):
            return 0
    else:
        #FIXME Cannot currentlyt validate this
        pass

    return 1
